#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['snakipy']

package_data = \
{'': ['*']}

install_requires = \
['abc_algorithm', 'numpy >= 1.16.4', 'fire >= 0.2.1', 'tqdm >= 4.32.1']

entry_points = \
{'console_scripts': ['snake = snakipy.ui:entrypoint']}

setup(name='snakipy',
      version='0.0.4',
      description='Snake',
      author='Gabriel Kabbe',
      author_email='gabriel.kabbe@mail.de',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
     )
